
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; MODULE      : translate.scm
;; DESCRIPTION : support for automatic translations
;; COPYRIGHT   : (C) 2006  Joris van der Hoeven
;;
;; This software falls under the GNU general public license and comes WITHOUT
;; ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for details.
;; If you don't have this file, write to the Free Software Foundation, Inc.,
;; 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(texmacs-module (doc translate)
  (:use (convert tools tmconcat)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Determine parts of the document which should be translated and
;; decompose these parts into sentences
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (translate-break-sentences l)
  (cond ((null? l) '(()))
        ((and (in? (car l) '(#\. #\! #\?))
              (nnull? (cdr l)) (== (cadr l) #\space))
         (with r (translate-break-sentences (cddr l))
           (cons (list (car l)) r)))
        (else (with r (translate-break-sentences (cdr l))
                (cons (cons (car l) (car r)) (cdr r))))))

(define (translate-make-sentences-sub t)
  (cond ((== t "") '())
        ((string? t)
         (with r (translate-break-sentences (string->list t))
           (map list->string r)))
        ((func? t 'concat)
         (apply append (map translate-make-sentences-sub (cdr t))))
	((== t '(item)) '())
        (else (list t))))

(define (translate-merge-sentences l)
  (cond ((null? l) '(()))
        ((with s (car l)
           (and (string? s) (!= s "")
                (in? (string-ref s (- (string-length s) 1)) '(#\. #\! #\?))))
         (cons (list (car l)) (translate-merge-sentences (cdr l))))
        (else (with r (translate-merge-sentences (cdr l))
                (cons (cons (car l) (car r)) (cdr r))))))

(define (translate-make-sentences t)
  (let* ((l (translate-make-sentences-sub t))
         (m (translate-merge-sentences l)))
    (map (lambda (x) (apply tmconcat* x)) m)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Determine matching phrases between document and its translation
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (translate-match-sentences l1 l2)
  (if (or (null? l1) (!= (length l1) (length l2))) '()
      (let ((t1 (car l1))
            (t2 (car l2))
            (r (translate-match-sentences (cdr l1) (cdr l2))))
        (cond ((or (== t1 "") (== t2 "")) r)
              ((and (or (string? t1) (func? t1 'concat))
		    (or (string? t2) (func? t2 'concat)))
               (cons (list t1 t2) r))
              ((and (nstring? t1) (not (func? t1 'concat))
                    (nstring? t2) (not (func? t2 'concat)))
               (append (translate-match t1 t2) r))
              (else r)))))

(tm-define (translate-match t1 t2)
  (cond ((== t1 t2) '())
	((and (or (string? t1) (func? t1 'concat))
	      (or (string? t2) (func? t2 'concat)))
	 (translate-match-sentences (translate-make-sentences t1)
				    (translate-make-sentences t2)))
        ((and (list? t1) (list? t2)
	      (== (car t1) (car t2))
	      (= (length t1) (length t2)))
         (with r (map translate-match (cdr t1) (cdr t2))
           (apply append r)))
        (else
	 ;;(display* "Dropped: " t1 "\n")
	 ;;(display* "-> " t2 "\n")
	 '())))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Flatten the list of matching sentences to a list of matching strings
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (translate-flatten-number l nr)
  (if (null? l) l
      (cons (cons (car l) nr)
            (translate-flatten-number (cdr l) (+ nr 1)))))

(define (translate-flatten-subst l a)
  (if (null? l) l
      (with r (translate-flatten-subst (cdr l) a)
        (if (string? (car l)) (cons (car l) r)
            (with nr (assoc-ref a (car l))
              (if (not nr) r
                  (cons (string-append "#" (number->string nr)) r)))))))

(define (translate-flatten-sub t1 t2)
  (cond ((and (string? t1) (string? t2)) (list t1 t2))
        ((and (func? t1 'concat) (func? t2 'concat))
         (let* ((l1 (list-filter (cdr t1) nstring?))
                (l2 (list-filter (cdr t2) nstring?)))
           (if (not (list-permutation? l1 l2)) #f
               (let* ((a (translate-flatten-number l1 1))
                      (r1 (translate-flatten-subst (cdr t1) a))
                      (r2 (translate-flatten-subst (cdr t2) a)))
                 (list (apply string-append r1)
                       (apply string-append r2))))))
        (else #f)))

(tm-define (translate-flatten l)
  (if (null? l) l
      (with (t1 t2) (car l)
        (let* ((x (translate-flatten-sub t1 t2))
               (r (translate-flatten (cdr l))))
          (if x (cons x r) r)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Build table with statistics on the number of occurrences of each word
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(tm-define (cdr>=? x y)
  (>= (cdr x) (cdr y)))

(define (word-list-sub l)
  (if (null? l) '(())
      (let* ((c (car l))
             (r (word-list-sub (cdr l))))
        (if (or (>= (char->integer c) 128) (char-alphabetic? c))
            (cons (cons c (car r)) (cdr r))
            (cons '() r)))))

(define (word-list s)
  (let* ((s2 (locase-all s))
         (l1 (string->list s2))
         (l2 (word-list-sub l1))
         (l3 (list-filter l2 nnull?)))
    (map list->string l3)))

(define (ahash-add! h x p)
  (if (not (ahash-ref h x)) (ahash-set! h x 0))
  (ahash-set! h x (+ (ahash-ref h x) p)))

(define (word-count-table-sub t h)
  (if (tree-atomic? t)
      (with l (word-list (tree->string t))
	(for-each (lambda (x) (ahash-add! h x 1)) l))
      (with l (tree-accessible-children t)
	(for-each (lambda (x) (word-count-table-sub x h)) l))))

(tm-define (word-count-table t)
  (with h (make-ahash-table)
    (word-count-table-sub t h)
    (let* ((l1 (ahash-table->list h)))
      (sort l1 cdr>=?))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Find plausible roots of words
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(tm-define (car-string<=? x y)
  (string<=? (car x) (car y)))

(define (longest-common-sublist l1 l2)
  (cond ((or (null? l1) (null? l2)) '())
	((== (car l1) (car l2))
	 (cons (car l1) (longest-common-sublist (cdr l1) (cdr l2))))
	(else '())))

(define (longest-common-substring s1 s2)
  (list->string (longest-common-sublist (string->list s1)
					(string->list s2))))

(define (word-longest-common-prefix l)
  (cond ((null? l) "")
	((null? (cdr l)) (car l))
	(else (longest-common-substring
	       (car l) (word-longest-common-prefix (cdr l))))))

(define (word-similar-sub l)
  (if (or (null? l) (null? (cdr l))) l
      (let* ((r1 (word-similar-sub (cdr l)))
	     (r2 (cons (caar l) (car r1)))
	     (pr (word-longest-common-prefix r2))
	     (ml (apply max (map string-length r2))))
	(if (<= ml (+ (string-length pr) 3))
	    (cons r2 (cdr r1))
	    (cons (car l) r1)))))

(tm-define (word-similar-table l)
  (with r (sort l string<=?)
    (word-similar-sub (map list r))))

(define (word-get-suffixes l)
  (with pr (word-longest-common-prefix l)
    (map (lambda (s) (string-tail s (string-length pr))) l)))

(define (list-count l)
  (let* ((t (make-ahash-table))
	 (insert (lambda (x) (ahash-add! t x 1))))
    (for-each insert l)
    (ahash-table->list t)))

(tm-define (word-suffix-table l)
  (let* ((l0 (word-similar-table l))
	 (l1 (apply append (map word-get-suffixes l0)))
	 (l2 (list-filter l1 (lambda (x) (!= x ""))))
	 (l3 (sort (list-count l2) cdr>=?))
	 (l4 (list-head l3 10))) ;; a bit heuristic
    (map car l4)))

(tm-define (word-remove-suffixes l suf)
  (let* ((pre (word-longest-common-prefix l))
	 (wsl (word-get-suffixes l))
	 (wws (map cons l wsl))
	 (fun (lambda (x) (cons (car x) (if (in? (cdr x) suf) pre (car x)))))
	 (r (map fun wws)))
    r))

(tm-define (word-root-table l)
  ;; FIXME: we should work with the original list and try to find pairs
  ;; of words which simplify to shortest common roots
  (let* ((sim (word-similar-table l))
	 (suf (word-suffix-table l))
	 (r (map (lambda (x) (word-remove-suffixes x suf)) sim)))
    (apply append r)))

(define (word-root-stats l t)
  ;; count nr of occurrences for roots instead of words themselves
  (with h (make-ahash-table)
    (for-each (lambda (x)
		(with r (ahash-ref t (car x))
		  (if r (ahash-add! h r (cdr x)))))
	      l)
    (ahash-table->list h)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Build table with statistics on plausible matching words
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define src-roots (make-ahash-table))
(define dest-roots (make-ahash-table))

(define (ahash-ref* t x z)
  (with y (ahash-ref t x)
    (if y y z)))

(define (word-map-table-sub l1 l2 t add)
  (cond ((string? l1)
         (when (nnull? l2)
	   (let* ((w1 (ahash-ref* src-roots l1 l1))
		  (w2 (ahash-ref* dest-roots (car l2) (car l2))))
	     (if (not (ahash-ref t w1))
		 (ahash-set! t w1 (make-ahash-table)))
	     (with u (ahash-ref t w1)
	       (ahash-add! u w2 add))
	     (word-map-table-sub l1 (cdr l2) t add))))
        ((null? l1) (noop))
        (else
         (word-map-table-sub (car l1) l2 t add)
         (word-map-table-sub (cdr l1) l2 t add))))

(tm-define (word-map-table l roots1 roots2)
  (set! src-roots roots1)
  (set! dest-roots roots2)
  (with t (make-ahash-table)
    (with f (lambda (x)
              (let* ((l1 (word-list (car x)))
                     (l2 (word-list (cadr x)))
                     (w (length l2))
                     (add (if (= w 0) 0 (/ 1 w))))
                (word-map-table-sub l1 l2 t add)))
      (for-each f l)
      (let* ((l1 (ahash-table->list t))
             (g (lambda (t) (list-sort (ahash-table->list t) cdr>=?)))
             (l2 (map (lambda (x) (cons (car x) (g (cdr x)))) l1)))
        l2))))

(define (assoc-index l x n)
  (cond ((null? l) n)
	((== (caar l) x) n)
	(else (assoc-index (cdr l) x (+ n 1)))))

(define (word-map-filter-sub w1 w2 t1 t2 ws1 ws2)
  ;;(display* w1 " -> " w2 "\n")
  (let* ((l1 (ahash-ref* t1 w1 '()))
	 ;;(d1 (display* "d1  = " l1 "\n"))
	 (l2 (ahash-ref* t2 w2 '()))
	 ;;(d2 (display* "d2  = " l2 "\n"))
	 (nr1 (assoc-index l1 w2 1))
	 ;;(d3 (display* "nr1 = " nr1 "\n"))
	 (nr2 (assoc-index l2 w1 1))
	 ;;(d4 (display* "nr2 = " nr2 "\n"))
	 (occ1 (ahash-ref* ws1 w1 0.1))
	 ;;(d5 (display* "occ1= " occ1 "\n"))
	 (occ2 (ahash-ref* ws2 w2 0.1))
	 ;;(d6 (display* "occ2= " occ2 "\n"))
	 (nr3 (max (/ occ1 occ2) (/ occ2 occ1)))
	 ;;(d7 (display* "nr3 = " nr3 "\n"))
	 (score (/ 100 (* nr1 nr2 nr3))))
    ;;(display* "score: " score "\n")
    (cons w2 score)))

(tm-define (word-map-filter l1 l2 s1 s2)
  (let* ((t1 (list->ahash-table l1))
	 (t2 (list->ahash-table l2))
	 (ws1 (list->ahash-table s1))
	 (ws2 (list->ahash-table s2)))
    (map (lambda (l)
	   (let* ((w1 (car l))
		  (ll (cdr l))
		  (rr (map (lambda (p)
			     (word-map-filter-sub w1 (car p) t1 t2 ws1 ws2))
			   ll))
		  (ss (sort rr cdr>=?)))
	     (cons w1 (list-head ss (min (length ss) 3)))))
	 l1)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Routines on files
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (url->list u)
  (cond ((url-none? u) '())
	((url-or? u)
	 (append (url->list (url-ref u 1))
		 (url->list (url-ref u 2))))
	(else (list u))))

(define (translate-process u)
  (let* ((en (url-glue (url-unglue u 6) ".en.tm"))
	 (d (display* en " -> " u "\n"))
	 (t1 (tree->stree (tree-load-inclusion en)))
	 (t2 (tree->stree (tree-load-inclusion u))))
    (translate-flatten (translate-match t1 t2))))

(tm-define (translate-pairs dir lancode)
  (let* ((u1 (url-append dir (url-any)))
	 (u2 (url-expand (url-complete u1 "dr")))
	 (u3 (url-append u2 (url-wildcard (string-append "*." lancode ".tm"))))
	 (u4 (url-expand (url-complete u3 "fr"))))
    (apply append (map (lambda (x) (translate-process x)) (url->list u4)))))

(define-public (test-translate)
  (translate-pairs "/home/vdhoeven/cvs/doc/main" "de"))

(tm-define (word-stats dir lancode)
  (let* ((u1 (url-append dir (url-any)))
	 (u2 (url-expand (url-complete u1 "dr")))
	 (u3 (url-append u2 (url-wildcard (string-append "*." lancode ".tm"))))
	 (u4 (url-expand (url-complete u3 "fr")))
	 (tl (map tree-load-inclusion (url->list u4))))
    (word-count-table (tm->tree `(document ,@tl)))))

(define-public (test-words)
  (word-stats "/home/vdhoeven/cvs/doc/main" "de"))

(tm-define (word-map dir lan)
  (let* ((s1 (word-stats dir "en"))
	 (s2 (word-stats dir lan))
	 (r1 (list->ahash-table (word-root-table (map car s1))))
	 (r2 (list->ahash-table (word-root-table (map car s2))))
	 (l1 (translate-pairs dir lan))
	 ;;(d1 (display* "l1= " l1 "\n"))
	 (l2 (map (lambda (x) (list (cadr x) (car x))) l1))
	 ;;(d2 (display* "l2= " l2 "\n"))
	 (m1 (word-map-table l1 r1 r2))
	 ;;(m1 (display* "m1= " m1 "\n"))
	 (m2 (word-map-table l2 r2 r1))
	 (ws1 (word-root-stats s1 r1))
	 (ws2 (word-root-stats s2 r2)))
    (let* ((M1 (word-map-filter m1 m2 ws1 ws2))
	   (M2 (word-map-filter m2 m1 ws2 ws1)))
      (word-map-filter M1 M2 ws1 ws2))))

(define-public (test-word-map)
  (word-map "/home/vdhoeven/cvs/doc/main" "es"))
